/**
 * @file Config.cpp
 * @author apio (cloudapio.eu)
 * @brief Configuration file parsing.
 *
 * @copyright Copyright (c) 2024, the Luna authors.
 *
 */

#include <os/Config.h>
#include <stdlib.h>

namespace os
{
    Result<OwnedPtr<ConfigFile>> ConfigFile::open(os::Path path)
    {
        auto config = TRY(make_owned<ConfigFile>());

        config->m_path = path;

        auto file = TRY(os::File::open(path, os::File::ReadOnly));

        while (true)
        {
            auto line = TRY(file->read_line());
            if (line.is_empty()) break;

            line.trim("\n");
            if (line.is_empty()) continue;

            auto parts = TRY(line.split_once('='));
            if (parts.size() < 2 || parts[0].is_empty() || parts[1].is_empty()) return err(EINVAL);

            auto key = TRY(RefString::from_string(parts[0]));
            auto value = TRY(RefString::from_string(parts[1]));

            config->m_data.try_set(key, value);
        }

        return config;
    }

    Option<StringView> ConfigFile::read_string(StringView key)
    {
        auto string = RefString::from_string_view(key).release_value();
        auto* value = m_data.try_get_ref(string);

        if (!value) return {};

        return value->view();
    }

    StringView ConfigFile::read_string_or(StringView key, StringView default_value)
    {
        return read_string(key).value_or(default_value);
    }

    Option<i64> ConfigFile::read_number(StringView key)
    {
        auto maybe_string = read_string(key);

        if (!maybe_string.has_value()) return {};

        auto string = maybe_string.value();

        char* endptr;

        i64 result = strtol(string.chars(), &endptr, 10);

        if (string.chars() == endptr) return {};

        return result;
    }

    i64 ConfigFile::read_number_or(StringView key, i64 default_value)
    {
        return read_number(key).value_or(default_value);
    }

    Option<bool> ConfigFile::read_boolean(StringView key)
    {
        auto maybe_string = read_string(key);

        if (!maybe_string.has_value()) return {};

        auto string = maybe_string.value();

        if (string == "true" || string.to_uint().value_or(0) == 1) return true;
        if (string == "false" || string.to_uint().value_or(1) == 0) return false;

        return {};
    }

    bool ConfigFile::read_boolean_or(StringView key, bool default_value)
    {
        return read_boolean(key).value_or(default_value);
    }

    Result<void> ConfigFile::write_string(StringView key, StringView value)
    {
        auto key_string = TRY(RefString::from_string_view(key));
        auto value_string = TRY(RefString::from_string_view(value));

        auto worked = TRY(m_data.try_set(key_string, value_string));
        if (!worked)
        {
            m_data.try_remove(key_string);
            check(TRY(m_data.try_set(key_string, value_string)));
        }

        return {};
    }

    Result<void> ConfigFile::write_number(StringView key, int value)
    {
        auto value_string = TRY(String::format("%d"_sv, value));

        return write_string(key, value_string.view());
    }

    Result<void> ConfigFile::write_boolean(StringView key, bool value)
    {
        StringView value_string = value ? "true" : "false";

        return write_string(key, value_string);
    }

    Result<void> ConfigFile::sync()
    {
        auto file = TRY(os::File::open(m_path, os::File::WriteOnly));

        for (auto& pair : m_data)
        {
            file->write_formatted("%s=%s\n", pair.key.chars(), pair.value.release_value().chars());
        }

        return {};
    }
}
